"""
Notes on this Provider

1. Make sure the NIOS Version support WAPI 2.6.1
2. Have a valid Certificate from a public CA installed at the Infoblox

Commandline examples:
1. all parameters given:
lexicon infoblox --ib-host myib.mydomain.tld --auth-user {username} --auth-psw {passwordd}
    --ib-view default create test.local A --content 10.10.10.11 --name lexicon1
2. Parameters mixed with ENV
LEXICON_INFOBLOX_AUTH_USER={user} LEXICON_INFOBLOX_AUTH_PSW={password} lexicon infoblox
    --ib-host myib.mydomain.tld --ib-view default create test.local A
    --content 10.10.10.11 --name lexicon1

"""
from __future__ import absolute_import

import logging
import json
import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['test.local.']
# SOA and NS are not record types itself, but rather auto created depending on the Zone settings.
# Either Primary Grid Member directly assigned to the zone, or through a ns_group
# skipping for now
'''
Dictionary to map the record Type to different Infoblox specific values
IB_TYPE2CONTENT = {
    'type':['WAPI atribute field content','WAPI base URL per type','WAPI returnfields'],
}
'''
IB_TYPE2CONTENT = {
    'A':['ipv4addr', 'record:a', ',ttl,use_ttl'],
    'AAAA':['ipv6addr', 'record:aaaa', ',ttl,use_ttl'],
    'CNAME':['canonical', 'record:cname', ',ttl,use_ttl'],
    'MX':['mail_exchanger', 'record:mx', ',ttl,use_ttl'],
    #'NS':['nameserver','record:ns',''],
    #'SOA':['nameserver','record:ns',''],
    'TXT':['text', 'record:txt', ',ttl,use_ttl'],
    'SRV':['target', 'record:srv', ',ttl,use_ttl'],
}

def provider_parser(subparser):
    """Configure provider parser for Infoblox"""
    subparser.add_argument('--auth-user', help='specify the user to access the Infoblox WAPI')
    subparser.add_argument('--auth-psw', help='specify the password to access the Infoblox WAPI')
    subparser.add_argument('--ib-view', default='default',
                           help='specify DNS View to manage at the Infoblox')
    subparser.add_argument('--ib-host', help='specify Infoblox Host exposing the WAPI')

class Provider(BaseProvider):
    """Provider class for Infoblox"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        # In Case .local. Domains are used, ignore the tldextract in client.py
        # this will be used in test_infoblox.py as well
        if self.config.resolve('lexicon:domain') in NAMESERVER_DOMAINS[0]:
            self.domain = NAMESERVER_DOMAINS[0].rstrip('.')
        self.domain_id = None
        self.version_id = None
        self.view = self._get_provider_option('ib_view')
        self.session = requests.session()
        self.session.auth = (self._get_provider_option('auth_user'),
                             self._get_provider_option('auth_psw'))
        self.version = '2.6.1' #WAPI version supported by NIOS 8.3 and above
        self.session.headers.update({'Content-Type' : 'application/json'})
        self.api_endpoint = ('https://{0}/wapi/v{1}/'
                             .format(self._get_provider_option('ib_host'), self.version))

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def _authenticate(self):
        response = self.session.get('{0}zone_auth?fqdn={1}&view={2}'
                                    .format(self.api_endpoint, self.domain, self.view))
        domains = response.json()
        try:
            self.domain_id = domains[0]['_ref']
        except IndexError:
            LOGGER.error('Domain %s not found in view', self.domain)
            raise Exception('Domain {0} not found in view'.format(self.domain))

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        if name:
            name = self._fqdn_name(name)
        else:
            raise Exception('Name not specified, no FQDN could be build')
        # Find existing records for all types
        existing = []
        for rrtype in IB_TYPE2CONTENT:
            existing = existing + self._list_records(rtype=rrtype, name=name)
        # we don't want to delete all existing A,AAAA,TXT,SRV,MX,NS records
        # which can not co-exist with CNAMEs
        if any(d['type'] == rtype for d in existing):
            # no conflict in types
            if any(d['content'] == content and d['type'] == rtype for d in existing):
                # already exists
                return True
            if rtype == 'CNAME':
                # we found a CNAME entry; we update it with the new target
                return self._update_record_internal(existing[0]['id'], rtype, name, content)
            return self._create_record_internal(rtype, name, content)
        if any(d['type'] == 'CNAME' and rtype != 'CNAME' for d in existing):
            # we found a CNAME entry; we can delete it and create the requested type
            if self._delete_record(identifier=existing[0]['id']):
                return self._create_record_internal(rtype, name, content)
            LOGGER.error('Deleting record failed for:%s', existing[0]['id'])
            return False
        if any(d['type'] != 'CNAME' and rtype == 'CNAME' for d in existing):
            LOGGER.error('CNAME requested, but other records already exists')
            raise Exception('CNAME requested, but other records already exists')
        return self._create_record_internal(rtype, name, content)

    def _create_record_internal(self, rtype, name, content):
        if name:
            name = self._fqdn_name(name)
        else:
            raise Exception('Name not specified, no FQDN could be build')
        if rtype.upper() in IB_TYPE2CONTENT:
            uri = '{0}{1}'.format(self.api_endpoint,
                                  IB_TYPE2CONTENT[rtype.upper()][1],
                                  )
            payload = self._generate_payload(rtype, name, content)
            try:
                response = self.session.post(uri, data=json.dumps(payload))
            except:
                LOGGER.error('Connection Error during create')
                raise Exception('Connection Error')
            if response.status_code == 201:
                return True
            return False
        raise Exception('RR Type not supported by Infoblox')
    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        # Infoblox stores entries based on their type, if type is not specified look up all types
        if not rtype:
            records = []
            for one_rtype in IB_TYPE2CONTENT:
                records = records + self._list_records_internal(
                    rtype=one_rtype, name=name, content=content)
            return records
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(self, rtype=None, name=None, content=None):
        # infoblox expect the fullname when looking it up without trailing dot
        if name:
            name = self._fqdn_name(name)
        else:
            name = self.domain
        records = []
        if rtype.upper() in IB_TYPE2CONTENT:
            uri = ('{0}{1}?name={2}&view={3}&_return_fields={4}{5}'
                   .format(self.api_endpoint,
                           IB_TYPE2CONTENT[rtype.upper()][1],
                           name,
                           self.view,
                           IB_TYPE2CONTENT[rtype.upper()][0],
                           IB_TYPE2CONTENT[rtype.upper()][2]))
            try:
                response = self.session.get(uri)
            except:
                raise Exception('Connection Error')

            results = response.json()
            for result in results:
                if content:
                    #return exact match
                    if str(content).lower() == str(result[IB_TYPE2CONTENT[rtype][0]]).lower():
                        if result['use_ttl']:
                            record = {
                                'type': rtype,
                                'name': name,
                                'ttl': result['ttl'],
                                'content': result[IB_TYPE2CONTENT[rtype][0]],
                                'id' : result['_ref'],
                            }
                        else:
                            record = {
                                'type': rtype,
                                'name': name,
                                'ttl': 3600, # replace by Infoblox default TTL
                                'content': result[IB_TYPE2CONTENT[rtype][0]],
                                'id' : result['_ref'],
                            }
                        records.append(record)
                else:
                    #return any record
                    if result['use_ttl']:
                        record = {
                            'type': rtype,
                            'name': name,
                            'ttl': result['ttl'],
                            'content': result[IB_TYPE2CONTENT[rtype][0]],
                            'id' : result['_ref'],
                        }
                    else:
                        record = {
                            'type': rtype,
                            'name': name,
                            'ttl': 3600, # replace by Infoblox default TTL
                            'content': result[IB_TYPE2CONTENT[rtype][0]],
                            'id' : result['_ref'],
                        }
                    records.append(record)
            return records
        return records

    # Update a record. Identifier must be specified.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            return self._update_record_internal(identifier=identifier,
                                                rtype=rtype, name=name, content=content)
        success = []
        for record in self._list_records(rtype, name, content):
            success.append(self._update_record_internal(identifier=record['id'],
                                                        rtype=rtype, name=name, content=content))
        if False in success:
            return False
        return True

    def _update_record_internal(self, identifier, rtype=None, name=None, content=None):
        uri = '{0}{1}'.format(self.api_endpoint, identifier)
        payload = self._generate_payload(rtype, name, content)
        # remove view and name from the payload, as they can not be updated
        del payload['view']
        del payload['name']
        try:
            response = self.session.put(uri, data=json.dumps(payload))
        except:
            LOGGER.error('Connection Error during create')
            raise Exception('Connection Error')
        if response.status_code == 200:
            return True
        return False

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Infoblox Object Identifier example:
        # record:cname/ZG5zLmJpbmRfY25hbWUkLjIzLmRlLm1naS5sZXhpY29uNA:lexicon.domain.tld/view
        if identifier:
            return self._delete_record_internal(identifier)
        success = []
        for record in self._list_records(rtype, name, content):
            success.append(self._delete_record_internal(record['id']))
        if False in success:
            return False
        return True

    def _delete_record_internal(self, identifier=None):
        if identifier:
            uri = '{0}{1}'.format(self.api_endpoint,
                                  identifier,
                                  )
            try:
                response = self.session.delete(uri)
            except:
                LOGGER.error('Connection Error during create')
                raise Exception('Connection Error')
            if response.status_code == 200:
                return True
            return False
        return False

    # Helpers
    def _generate_payload(self, rtype, name, content):
        # build full object as required by Infoblox WAPI
        payload = {}
        payload['view'] = self.view
        payload['name'] = name
        if rtype.upper() == 'TXT':
            payload[IB_TYPE2CONTENT[rtype.upper()][0]] = '"' + content + '"'
        else:
            payload[IB_TYPE2CONTENT[rtype.upper()][0]] = content
        if self._get_lexicon_option('ttl'):
            payload['ttl'] = self._get_lexicon_option('ttl')
            payload['use_ttl'] = True
        # MX and SRV have additional fields
        if rtype.upper() == 'MX':
            payload['preference'] = self._get_lexicon_option('priority')
        elif rtype.upper() == 'SRV':
            payload['priority'] = self._get_lexicon_option('priority')
            payload['weight'] = self._get_provider_option('ib_weight')
            payload['port'] = self._get_provider_option('ib_port')
        return payload

    def _fqdn_name(self, record_name):
        record_name = record_name.rstrip('.') # strip trailing period from fqdn if present
        # check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = "{0}.{1}".format(record_name, self.domain)
        return "{0}".format(record_name) #return the fqdn name without trailing dot

    def _request(self, action='GET', url='/', data=None, query_params=None):
        # Helper _request is not used in Infoflox provider
        pass



"""Module provider for DNS Simple"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['dnsimple.com']


def provider_parser(subparser):
    """Configure provider parser for DNS Simple"""
    subparser.add_argument(
        "--auth-token", help="specify api token for authentication")
    subparser.add_argument(
        "--auth-username", help="specify email address for authentication")
    subparser.add_argument(
        "--auth-password", help="specify password for authentication")
    subparser.add_argument(
        "--auth-2fa",
        help="specify two-factor auth token (OTP) to use with email/password authentication")


class Provider(BaseProvider):
    """Provider class for DNS Simple"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.account_id = None
        self.api_endpoint = self._get_provider_option(
            'api_endpoint') or 'https://api.dnsimple.com/v2'

    def _authenticate(self):

        payload = self._get('/accounts')

        if not payload[0]['id']:
            raise Exception('No account id found')

        for account in payload:
            dompayload = self._get(
                '/{0}/domains'.format(account['id']), query_params={'name_like': self.domain})
            if dompayload and dompayload[0]['id']:
                self.account_id = account['id']
                self.domain_id = dompayload[0]['id']

        if not self.account_id:
            raise Exception('No domain found like {}'.format(self.domain))

    # Create record. If record already exists with the same content, do nothing

    def _create_record(self, rtype, name, content):
        # check if record already exists
        existing_records = self._list_records(rtype, name, content)
        if len(existing_records) == 1:
            return True

        record = {
            'type': rtype,
            'name': self._relative_name(name),
            'content': content
        }
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            record['priority'] = self._get_lexicon_option('priority')
        if self._get_provider_option('regions'):
            record['regions'] = self._get_provider_option('regions')

        payload = self._post(
            '{0}/zones/{1}/records'.format(self.account_id, self.domain), record)

        LOGGER.debug('create_record: %s', 'id' in payload)
        return 'id' in payload

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_query = {}
        if rtype:
            filter_query['type'] = rtype
        if name:
            filter_query['name'] = self._relative_name(name)
        payload = self._get(
            '/{0}/zones/{1}/records'.format(self.account_id, self.domain),
            query_params=filter_query)

        records = []
        for record in payload:
            processed_record = {
                'type': record['type'],
                'name': '{}'.format(
                    self.domain) if record['name'] == "" else '{0}.{1}'.format(
                        record['name'],
                        self.domain),
                'ttl': record['ttl'],
                'content': record['content'],
                'id': record['id']}
            if record['priority']:
                processed_record['priority'] = record['priority']
            records.append(processed_record)

        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {}

        if identifier is None:
            records = self._list_records(rtype, name, content)
            identifiers = [record["id"] for record in records]
        else:
            identifiers = [identifier]

        if name:
            data['name'] = self._relative_name(name)
        if content:
            data['content'] = content
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            data['priority'] = self._get_lexicon_option('priority')
        if self._get_provider_option('regions'):
            data['regions'] = self._get_provider_option('regions')

        for one_identifier in identifiers:
            self._patch('/{0}/zones/{1}/records/{2}'
                        .format(self.account_id, self.domain, one_identifier), data)
            LOGGER.debug('update_record: %s', one_identifier)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self._delete(
                '/{0}/zones/{1}/records/{2}'.format(self.account_id, self.domain, record_id))

        # is always True at this point; if a non 2xx response is returned, an error is raised.
        LOGGER.debug('delete_record: True')
        return True

    # Helpers

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        default_auth = None

        if self._get_provider_option('auth_token'):
            default_headers['Authorization'] = "Bearer {0}".format(
                self._get_provider_option('auth_token'))
        elif (self._get_provider_option('auth_username')
              and self._get_provider_option('auth_password')):
            default_auth = (self._get_provider_option(
                'auth_username'), self._get_provider_option('auth_password'))
            if self._get_provider_option('auth_2fa'):
                default_headers['X-Dnsimple-OTP'] = self._get_provider_option(
                    'auth_2fa')
        else:
            raise Exception('No valid authentication mechanism found')

        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers,
                                    auth=default_auth)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if response.text and response.json()['data'] is None:
            raise Exception('No data returned')

        return response.json()['data'] if response.text else None

    def _patch(self, url='/', data=None, query_params=None):
        return self._request('PATCH', url, data=data, query_params=query_params)



# pylint: disable=missing-docstring
import os

import pytest

import lexicon.client
from lexicon.config import ConfigResolver


def test_client_basic_init():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == options['domain']
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_legacy_init():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == options['domain']
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_init_when_domain_includes_subdomain_should_strip():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'www.example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == 'example.com'
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_init_with_delegated_domain_name():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'www.sub.example.com',
        'delegated': 'sub',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == "sub.example.com"
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_init_with_delegated_domain_fqdn():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'www.sub.example.com',
        'delegated': 'sub.example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == "sub.example.com"
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_init_with_same_delegated_domain_fqdn():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'www.example.com',
        'delegated': 'example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == "example.com"
    assert client.config.resolve('lexicon:type') == options['type']


def test_client_init_when_missing_provider_should_fail():
    options = {
        'action': 'list',
        'domain': 'example.com',
        'type': 'TXT'
    }
    with pytest.raises(AttributeError):
        lexicon.client.Client(options)


def test_client_init_when_missing_action_should_fail():
    options = {
        'provider_name': 'base',
        'domain': 'example.com',
        'type': 'TXT'
    }
    with pytest.raises(AttributeError):
        lexicon.client.Client(options)


def test_client_init_when_missing_domain_should_fail():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'type': 'TXT'
    }
    with pytest.raises(AttributeError):
        lexicon.client.Client(options)


def test_client_init_when_missing_type_should_fail():
    options = {
        'provider_name': 'base',
        'action': 'list',
        'domain': 'example.com',
    }
    with pytest.raises(AttributeError):
        lexicon.client.Client(options)


def test_client_parse_env_with_no_keys_should_do_nothing(monkeypatch):
    if os.environ.get('LEXICON_CLOUDFLARE_TOKEN'):
        monkeypatch.delenv('LEXICON_CLOUDFLARE_TOKEN')
    if os.environ.get('LEXICON_CLOUDFLARE_USERNAME'):
        monkeypatch.delenv('LEXICON_CLOUDFLARE_USERNAME')
    options = {
        'provider_name': 'cloudflare',
        'action': 'list',
        'domain': 'www.example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == 'example.com'
    assert client.config.resolve('lexicon:type') == options['type']
    assert client.config.resolve('lexicon:cloudflare:auth_token') is None
    assert client.config.resolve('lexicon:cloudflare:auth_username') is None


def test_client_parse_env_with_auth_keys(monkeypatch):
    monkeypatch.setenv('LEXICON_CLOUDFLARE_TOKEN', 'test-token')
    monkeypatch.setenv('LEXICON_CLOUDFLARE_USERNAME',
                       'test-username@example.com')
    options = {
        'provider_name': 'cloudflare',
        'action': 'list',
        'domain': 'www.example.com',
        'type': 'TXT'
    }
    client = lexicon.client.Client(options)

    assert client.provider_name == options['provider_name']
    assert client.action == options['action']
    assert client.config.resolve('lexicon:domain') == 'example.com'
    assert client.config.resolve('lexicon:type') == options['type']
    assert client.config.resolve(
        'lexicon:cloudflare:auth_token') == 'test-token'
    assert client.config.resolve(
        'lexicon:cloudflare:auth_username') == 'test-username@example.com'

# TODO: add tests for Provider loading?



"""Integration tests for Goddady"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.godaddy import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class GoDaddyProviderTests(TestCase, IntegrationTests):
    """TestCase for Godaddy"""
    Provider = Provider
    provider_name = 'godaddy'
    domain = 'fullm3tal.online'

    def _filter_headers(self):
        return ['Authorization']



# """Test for route53 implementation of the interface."""
# import unittest
# import pytest
# from lexicon.providers.route53 import Provider
# from lexicon.tests.providers.integration_tests import IntegrationTests, provider_vcr
#
# class Route53ProviderTests(unittest.TestCase, IntegrationTests):
#     """Route53 Proivder Tests."""
#
#     Provider = Provider
#     provider_name = 'route53'
#     domain = 'capsulecd.com'
#
#     def _filter_headers(self):
#         """Sensitive headers to be filtered."""
#         return ['Authorization']
#
#     def test_provider_authenticate_private_zone_only(self):
#         with provider_vcr.use_cassette(super(Route53ProviderTests, self)
#               ._cassette_path('IntegrationTests/test_provider_authenticate.yaml'),
#                   filter_headers=super(Route53ProviderTests, self)._filter_headers(),
#                   filter_query_parameters=super(Route53ProviderTests, self).
#                       ._filter_query_parameters(),
#                   filter_post_data_parameters=super(Route53ProviderTests, self)
#                       ._filter_post_data_parameters()):
#             options = super(Route53ProviderTests, self)._test_options()
#             options['private_zone'] = 'true'
#             provider = self.Provider(options, super(Route53ProviderTests, self)
#                           ._test_engine_overrides())
#             with pytest.raises(Exception):
#                 provider.authenticate()
#
#     def test_provider_authenticate_private_zone_false(self):
#         with provider_vcr.use_cassette(super(Route53ProviderTests, self)
#           ._cassette_path('IntegrationTests/test_provider_authenticate.yaml'),
#               filter_headers=super(Route53ProviderTests, self)._filter_headers(),
#               filter_query_parameters=super(Route53ProviderTests, self)
#                   ._filter_query_parameters(),
#               filter_post_data_parameters=super(Route53ProviderTests, self)
#                   ._filter_post_data_parameters()):
#             options = super(Route53ProviderTests, self)._test_options()
#             options['private_zone'] = 'false'
#             provider = self.Provider(options, super(Route53ProviderTests, self)
#                               ._test_engine_overrides())
#             provider.authenticate()
#             assert provider.domain_id is not None
#
#     @pytest.mark.skip(reason="route 53 dns records don't have ids")
#     def test_provider_when_calling_delete_record_by_identifier_should_remove_record(self):
#         return
#
#     # TODO: the following skipped suite and fixtures should be enabled
#     @pytest.mark.skip(reason="new test, missing recording")
#     def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
#         return
#
#     @pytest.fixture(autouse=True)
#     def _skip_suite(self, request):  # pylint: disable=no-self-use
#         if request.node.get_closest_marker('ext_suite_1'):
#             pytest.skip('Skipping extended suite')



"""
A note about running these tests against the live environment

1.  Namecheap offers a Testing Sandbox for their API.
    You must create an account on `sandbox.namecheap.com`, then "apply" for API
    access. It is supposed to be issued automatically, but you will most-likely
    need to ask their support to approve API access on the test system. You can
    do this via their LiveChat

    https://www.namecheap.com/support/api/methods.aspx

2.  There are two variants of the test:
    * NamecheapProviderTests
    * NamecheapManagedProviderTests

    Namecheap acts a little differently for "owned" vs "managed" domains, so a
    secondary test run must be done.

To set enable live testing against the actual API:

* Create two (2) Sandbox accounts on Namecheap
* Using the First Account:
    *   Enable API access
    *   "Purchase" a first test domain for use on `NamecheapProviderTests`.
        This will be the env variable `LEXICON_NAMECHEAP_DOMAIN`
* Using the Second Account:
    *   "Purchase" a second test domain for use on `NamecheapManagedProviderTests`.
        This will be the env variable `LEXICON_NAMECHEAP_DOMAINMANAGED`
    *   Using the Namecheap dashboard, "share" the domain management with the
        first user.
*   The First account will then get an email to "accept" the domain management
    invitation from the Second account.
*   All API tests occure on the First account
*   Note: Namecheap's API requires the client's IP address to be whitelisted.

The required Environment Variables for a live test are:

    export LEXICON_NAMECHEAP_TOKEN={TOKEN}
    export LEXICON_NAMECHEAP_USERNAME={USERNAME}
    export LEXICON_NAMECHEAP_DOMAIN={DOMAIN_1}
    export LEXICON_NAMECHEAP_DOMAINMANAGED={DOMAIN_2}
"""
import os
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.namecheap import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class NamecheapProviderTests(TestCase, IntegrationTests):
    """TestCase for Namecheap"""
    Provider = Provider
    provider_name = 'namecheap'

    @property
    def domain(self):
        """
        this can be used to override the tests
            LEXICON_NAMECHEAP_DOMAIN
        """
        env_domain = os.environ.get('LEXICON_NAMECHEAP_DOMAIN', None)
        return env_domain or 'example-aptise.com'

    def _filter_query_parameters(self):
        return ['ApiKey', 'UserName', 'ApiUser']

    def _test_parameters_overrides(self):
        return {
            'auth_sandbox': True,
            'auth_client_ip': '127.0.0.1'
        }

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return


class NamecheapManagedProviderTests(NamecheapProviderTests):
    """
    The Namecheap API behaves differently for domains that are "Managed" by an
    account instead of "Owned" by the account.  Some endpoints won't work;
    others return different data.

    In orde to handle this, we run the tests on a second domain owned by another
    namecheap customer, but permissioned to this account.

    Note we define a `provider_variant`, which will change the cassette path.
    """
    provider_variant = 'managed'

    @property
    def domain(self):
        """
        this can be used to override the tests
            LEXICON_NAMECHEAP_DOMAINMANAGED
        """
        env_domain = os.environ.get('LEXICON_NAMECHEAP_DOMAINMANAGED', None)
        return env_domain or 'example-aptise-2.com'



"""Module provider for Hetzner"""
from __future__ import absolute_import
from __future__ import unicode_literals

from contextlib import contextmanager
import hashlib
import logging
import re
import time
import requests
from six import string_types
from urllib3.util.retry import Retry

# Due to optional requirement
try:
    from bs4 import BeautifulSoup
    import dns.resolver
    import dns.zone
except ImportError:
    pass

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = []


def provider_parser(subparser):
    """Configure a provider parser for Hetzner"""
    subparser.add_argument('--auth-account',
                           help='specify type of Hetzner account: by default Hetzner Robot '
                           '(robot) or Hetzner konsoleH (konsoleh)')
    subparser.add_argument('--auth-username', help='specify username of Hetzner account')
    subparser.add_argument('--auth-password', help='specify password of Hetzner account')
    subparser.add_argument('--linked',
                           help='if exists, uses linked CNAME as A|AAAA|TXT record name for edit '
                           'actions: by default (yes); Further restriction: Only enabled if '
                           'record name or raw FQDN record identifier \'type/name/content\' is '
                           'specified, and additionally for update actions the record name '
                           'remains the same',
                           default=str('yes'),
                           choices=['yes', 'no'])
    subparser.add_argument('--propagated',
                           help='waits until record is publicly propagated after succeeded '
                           'create|update actions: by default (yes)',
                           default=str('yes'),
                           choices=['yes', 'no'])
    subparser.add_argument('--latency',
                           help='specify latency, used during checks for publicly propagation '
                           'and additionally for Hetzner Robot after record edits: by default '
                           '30s (30)',
                           default=int(30),
                           type=int)


class Provider(BaseProvider):
    """
    Implements the Hetzner DNS Provider.
    There are two variants to manage DNS records on Hetzner: Hetzner Robot or
    Hetzner konsoleH. Both do not provide a common API, therefore this provider
    implements missing read and write methods in a generic way. For editing DNS
    records on Hetzner, this provider manipulates and replaces the whole DNS zone.
    Furthermore, there is no unique identifier to each record in the way that Lexicon
    expects, why this provider implements a pseudo-identifer based on the record type,
    name and content for use of the --identifier parameter. Supported identifier
    formats are:
        - hash  generated|verified by 'list' command; e.g. '30fa112'
        - raw   concatenation of the record type, name (FQDN) and content (if possible
                FQDN) with delimiter '/'; e.g. 'SRV/example.com./0 0 443 msx.example.com.'
                or 'TXT/example.com./challengetoken'
    Additional, this provider implements the option of replacing an A, AAAA or TXT record
    name with an existent linked CNAME for edit actions via the --linked parameter and
    the option of waiting until record is publicly propagated after succeeded create or
    update actions via the --propagated parameter. As further restriction, the use of a
    linked CNAME is only enabled if the record type & record name or the raw identifier are
    specified, and additionally for the update action the record name remains the same.
    """
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api = {
            'robot': {
                'endpoint': 'https://robot.your-server.de',
                'filter': [{'name': 'div', 'attrs': {'id': 'center_col'}}],
                'auth': {
                    'endpoint': 'https://accounts.hetzner.com',
                    'GET': {'url': '/login'},
                    'POST': {'url': '/login_check'},
                    'filter': [{'name': 'form', 'attrs': {'id': 'login-form'}}],
                    'user': '_username',
                    'pass': '_password'
                },
                'exit': {
                    'GET': {'url': '/login/logout/r/true'}
                },
                'domain_id': {
                    'GET': {'url': '/dns/index/page/<index>'},
                    'filter': [
                        {'name': 'div', 'attrs': {'id': 'center_col'}},
                        {'name': 'table', 'attrs': {'class': 'box_title'}}
                    ],
                    'domain': [{'name': 'td', 'attrs': {'class': 'title'}}],
                    'id': {'attr': 'onclick', 'regex': r'\'(\d+)\''}
                },
                'zone': {
                    'GET': [{'url': '/dns/update/id/<id>'}],
                    'POST': {'url': '/dns/update'},
                    'filter': [
                        {'name': 'div', 'attrs': {'id': 'center_col'}},
                        {'name': 'ul', 'attrs': {'class': 'error_list'}}
                    ],
                    'file': 'zonefile'
                }
            },
            'konsoleh': {
                'endpoint': 'https://konsoleh.your-server.de',
                'filter': [{'name': 'div', 'attrs': {'id': 'content'}}],
                'auth': {
                    'GET': {},
                    'POST': {'url': '/login.php'},
                    'filter': [{'name': 'form', 'attrs': {'id': 'loginform'}}],
                    'user': 'login_user_inputbox',
                    'pass': 'login_pass_inputbox'
                },
                'exit': {
                    'GET': {'url': '/logout.php'}
                },
                'domain_id': {
                    'GET': {'params': {'page': '<index>'}},
                    'filter': [
                        {'name': 'div', 'attrs': {'id': 'domainlist'}},
                        {'name': 'dl'},
                        {'name': 'a'}
                    ],
                    'domain': [{'name': 'strong'}],
                    'id': {'attr': 'href', 'regex': r'=(D\d+)'}
                },
                'zone': {
                    'GET': [
                        {'params': {'domain_number': '<id>'}},
                        {'url': '/dns.php', 'params': {'dnsaction2': 'editintextarea'}}
                    ],
                    'POST': {'url': '/dns.php'},
                    'filter': [
                        {'name': 'div', 'attrs': {'id': 'content'}},
                        {'name': 'div', 'attrs': {'class': 'error'}}
                    ],
                    'file': 'zone_file1'
                }
            }
        }
        self.session = None

        self.account = self._get_provider_option('auth_account')
        if self.account in (None, 'robot', 'konsoleh'):
            self.account = self.account if self.account else 'robot'
        else:
            LOGGER.error('Hetzner => Argument for --auth-account is invalid: \'%s\' '
                         '(choose from \'robot\' or \'konsoleh\')', self.account)
            raise AssertionError
        self.username = self._get_provider_option('auth_username')
        assert self.username is not None
        self.password = self._get_provider_option('auth_password')
        assert self.password is not None

    def _authenticate(self):
        """
        Connects to Hetzner account and returns, if authentification was
        successful and the domain or CNAME target is managed by this account.
        """
        with self._session(self.domain, get_zone=False):
            return True

    def _create_record(self, rtype, name, content):
        """
        Connects to Hetzner account, adds a new record to the zone and returns a
        boolean, if creation was successful or not. Needed record rtype, name and
        content for record to create.
        """
        with self._session(self.domain, self.domain_id) as ddata:
            # Validate method parameters
            if not rtype or not name or not content:
                LOGGER.warning('Hetzner => Record has no rtype|name|content specified')
                return False

            # Add record to zone
            name = ddata['cname'] if ddata['cname'] else self._fqdn_name(name)
            rrset = ddata['zone']['data'].get_rdataset(name, rdtype=rtype, create=True)
            for rdata in rrset:
                if self._convert_content(rtype, content) == rdata.to_text():
                    LOGGER.info('Hetzner => Record with content \'%s\' already exists',
                                content)
                    return True

            ttl = (rrset.ttl if rrset.ttl > 0
                   and rrset.ttl < self._get_lexicon_option('ttl')
                   else self._get_lexicon_option('ttl'))
            rdataset = dns.rdataset.from_text(rrset.rdclass, rrset.rdtype,
                                              ttl, self._convert_content(rtype, content))
            rrset.update(rdataset)
            # Post zone to Hetzner
            synced_change = self._post_zone(ddata['zone'])
            if synced_change:
                self._propagated_record(rtype, name, self._convert_content(rtype, content),
                                        ddata['nameservers'])
            return synced_change

    def _list_records(self, rtype=None, name=None, content=None):
        """
        Connects to Hetzner account and returns a list of records filtered by record
        rtype, name and content. The list is empty if no records found.
        """
        with self._session(self.domain, self.domain_id) as ddata:
            name = self._fqdn_name(name) if name else None
            return self._list_records_in_zone(ddata['zone']['data'], rtype, name, content)

    def _update_record(self, identifier=None, rtype=None, name=None, content=None):  # pylint: disable=too-many-locals,too-many-branches
        """
        Connects to Hetzner account, changes an existing record and returns a boolean,
        if update was successful or not. Needed identifier or rtype & name to lookup
        over all records of the zone for exactly one record to update.
        """
        with self._session(self.domain, self.domain_id) as ddata:
            # Validate method parameters
            if identifier:
                dtype, dname, dcontent = self._parse_identifier(identifier, ddata['zone']['data'])
                if dtype and dname and dcontent:
                    rtype = rtype if rtype else dtype
                    name = name if name else dname
                    content = content if content else dcontent
                else:
                    LOGGER.warning('Hetzner => Record with identifier \'%s\' does not exist',
                                   identifier)
                    return False

            elif rtype and name and content:
                dtype, dname, dcontent = rtype, name, None
            else:
                LOGGER.warning('Hetzner => Record has no rtype|name|content specified')
                return False

            dname = ddata['cname'] if ddata['cname'] else self._fqdn_name(dname)
            records = self._list_records_in_zone(ddata['zone']['data'], dtype, dname, dcontent)
            if len(records) == 1:
                # Remove record from zone
                rrset = ddata['zone']['data'].get_rdataset(records[0]['name'] + '.',
                                                           rdtype=records[0]['type'])
                rdatas = []
                for rdata in rrset:
                    if self._convert_content(records[0]['type'],
                                             records[0]['content']) != rdata.to_text():
                        rdatas.append(rdata.to_text())
                if rdatas:
                    rdataset = dns.rdataset.from_text_list(rrset.rdclass, rrset.rdtype,
                                                           records[0]['ttl'], rdatas)
                    ddata['zone']['data'].replace_rdataset(records[0]['name'] + '.', rdataset)
                else:
                    ddata['zone']['data'].delete_rdataset(records[0]['name'] + '.',
                                                          records[0]['type'])
                # Add record to zone
                name = ddata['cname'] if ddata['cname'] else self._fqdn_name(name)
                rrset = ddata['zone']['data'].get_rdataset(name, rdtype=rtype, create=True)
                synced_change = False
                for rdata in rrset:
                    if self._convert_content(rtype, content) == rdata.to_text():
                        LOGGER.info('Hetzner => Record with content \'%s\' already exists',
                                    content)
                        synced_change = True
                        break
                if not synced_change:
                    ttl = (rrset.ttl if rrset.ttl > 0
                           and rrset.ttl < self._get_lexicon_option('ttl')
                           else self._get_lexicon_option('ttl'))
                    rdataset = dns.rdataset.from_text(rrset.rdclass, rrset.rdtype, ttl,
                                                      self._convert_content(rtype, content))
                    rrset.update(rdataset)
                # Post zone to Hetzner
                synced_change = self._post_zone(ddata['zone'])
                if synced_change:
                    self._propagated_record(rtype, name, self._convert_content(rtype, content),
                                            ddata['nameservers'])
                return synced_change

            LOGGER.warning('Hetzner => Record lookup has not only one match')
            return False

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Connects to Hetzner account, removes an existing record from the zone and returns a
        boolean, if deletion was successful or not. Uses identifier or rtype, name & content to
        lookup over all records of the zone for one or more records to delete.
        """
        with self._session(self.domain, self.domain_id) as ddata:
            # Validate method parameters
            if identifier:
                rtype, name, content = self._parse_identifier(identifier, ddata['zone']['data'])
                if rtype is None or name is None or content is None:
                    LOGGER.info('Hetzner => Record with identifier \'%s\' does not exist',
                                identifier)
                    return True

            name = ddata['cname'] if ddata['cname'] else (self._fqdn_name(name) if name else None)
            records = self._list_records_in_zone(ddata['zone']['data'], rtype, name, content)
            if records:
                # Remove records from zone
                for record in records:
                    rrset = ddata['zone']['data'].get_rdataset(record['name'] + '.',
                                                               rdtype=record['type'])
                    rdatas = []
                    for rdata in rrset:
                        if self._convert_content(record['type'],
                                                 record['content']) != rdata.to_text():
                            rdatas.append(rdata.to_text())
                    if rdatas:
                        rdataset = dns.rdataset.from_text_list(rrset.rdclass, rrset.rdtype,
                                                               record['ttl'], rdatas)
                        ddata['zone']['data'].replace_rdataset(record['name'] + '.', rdataset)
                    else:
                        ddata['zone']['data'].delete_rdataset(record['name'] + '.', record['type'])
                # Post zone to Hetzner
                synced_change = self._post_zone(ddata['zone'])
                return synced_change

            LOGGER.info('Hetzner => Record lookup has no matches')
            return True

    ###############################################################################
    # Provider base helpers
    ###############################################################################

    @staticmethod
    def _create_identifier(rdtype, name, content):
        """
        Creates hashed identifier based on full qualified record type, name & content
        and returns hash.
        """
        sha256 = hashlib.sha256()
        sha256.update((rdtype + '/').encode('UTF-8'))
        sha256.update((name + '/').encode('UTF-8'))
        sha256.update(content.encode('UTF-8'))
        return sha256.hexdigest()[0:7]

    def _parse_identifier(self, identifier, zone=None):
        """
        Parses the record identifier and returns type, name & content of the associated record
        as tuple. The tuple is empty if no associated record found.
        """
        rdtype, name, content = None, None, None
        if len(identifier) > 7:
            parts = identifier.split('/')
            rdtype, name, content = parts[0], parts[1], '/'.join(parts[2:])
        else:
            records = self._list_records_in_zone(zone)
            for record in records:
                if record['id'] == identifier:
                    rdtype, name, content = record['type'], record['name'] + '.', record['content']
        return rdtype, name, content

    def _convert_content(self, rdtype, content):
        """
        Converts type dependent record content into well formed and fully qualified
        content for domain zone and returns content.
        """
        if rdtype == 'TXT':
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        if rdtype in ('CNAME', 'MX', 'NS', 'SRV'):
            if content[-1] != '.':
                content = self._fqdn_name(content)
        return content

    def _list_records_in_zone(self, zone, rdtype=None, name=None, content=None):
        """
        Iterates over all records of the zone and returns a list of records filtered
        by record type, name and content. The list is empty if no records found.
        """
        records = []
        rrsets = zone.iterate_rdatasets() if zone else []
        for rname, rdataset in rrsets:
            rtype = dns.rdatatype.to_text(rdataset.rdtype)
            if ((not rdtype or rdtype == rtype)
                    and (not name or name == rname.to_text())):
                for rdata in rdataset:
                    rdata = rdata.to_text()
                    if (not content or self._convert_content(rtype, content) == rdata):
                        raw_rdata = self._clean_TXT_record({'type': rtype,
                                                            'content': rdata})['content']
                        data = {
                            'type': rtype,
                            'name': rname.to_text(True),
                            'ttl': int(rdataset.ttl),
                            'content': raw_rdata,
                            'id': Provider._create_identifier(rtype, rname.to_text(), raw_rdata)
                        }
                        records.append(data)
        return records

    def _request(self, action='GET', url='/', data=None, query_params=None):
        """
        Requests to Hetzner by current session and returns the response.
        """
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = self.session.request(action, self.api[self.account]['endpoint'] + url,
                                        params=query_params, data=data)
        response.raise_for_status()
        return response

    ###############################################################################
    # Provider option helpers
    ###############################################################################

    @staticmethod
    def _dns_lookup(name, rdtype, nameservers=None):
        """
        Looks on specified or default system domain nameservers to resolve record type
        & name and returns record set. The record set is empty if no propagated
        record found.
        """
        rrset = dns.rrset.from_text(name, 0, 1, rdtype)
        try:
            resolver = dns.resolver.Resolver()
            resolver.lifetime = 1
            if nameservers:
                resolver.nameservers = nameservers
            rrset = resolver.query(name, rdtype)
            for rdata in rrset:
                LOGGER.debug('DNS Lookup => %s %s %s %s',
                             rrset.name.to_text(), dns.rdataclass.to_text(rrset.rdclass),
                             dns.rdatatype.to_text(rrset.rdtype), rdata.to_text())
        except dns.exception.DNSException as error:
            LOGGER.debug('DNS Lookup => %s', error)
        return rrset

    @staticmethod
    def _get_nameservers(domain):
        """
        Looks for domain nameservers and returns the IPs of the nameservers as a list.
        The list is empty, if no nameservers were found. Needed associated domain zone
        name for lookup.
        """
        nameservers = []
        rdtypes_ns = ['SOA', 'NS']
        rdtypes_ip = ['A', 'AAAA']
        for rdtype_ns in rdtypes_ns:
            for rdata_ns in Provider._dns_lookup(domain, rdtype_ns):
                for rdtype_ip in rdtypes_ip:
                    for rdata_ip in Provider._dns_lookup(rdata_ns.to_text().split(' ')[0],
                                                         rdtype_ip):
                        if rdata_ip.to_text() not in nameservers:
                            nameservers.append(rdata_ip.to_text())
        LOGGER.debug('DNS Lookup => %s IN NS %s', domain, ' '.join(nameservers))
        return nameservers

    @staticmethod
    def _get_dns_cname(name, link=False):
        """
        Looks for associated domain zone, nameservers and linked record name until no
        more linked record name was found for the given fully qualified record name or
        the CNAME lookup was disabled, and then returns the parameters as a tuple.
        """
        resolver = dns.resolver.Resolver()
        resolver.lifetime = 1
        domain = dns.resolver.zone_for_name(name, resolver=resolver).to_text(True)
        nameservers = Provider._get_nameservers(domain)
        cname = None
        links, max_links = 0, 5
        while link:
            if links >= max_links:
                LOGGER.error('Hetzner => Record %s has more than %d linked CNAME '
                             'records. Reduce the amount of CNAME links!',
                             name, max_links)
                raise AssertionError
            qname = cname if cname else name
            rrset = Provider._dns_lookup(qname, 'CNAME', nameservers)
            if rrset:
                links += 1
                cname = rrset[0].to_text()
                qdomain = dns.resolver.zone_for_name(cname, resolver=resolver).to_text(True)
                if domain != qdomain:
                    domain = qdomain
                    nameservers = Provider._get_nameservers(qdomain)
            else:
                link = False
        if cname:
            LOGGER.info('Hetzner => Record %s has CNAME %s', name, cname)
        return domain, nameservers, cname

    def _link_record(self):
        """
        Checks restrictions for use of CNAME lookup and returns a tuple of the
        fully qualified record name to lookup and a boolean, if a CNAME lookup
        should be done or not. The fully qualified record name is empty if no
        record name is specified by this provider.
        """
        action = self._get_lexicon_option('action')
        identifier = self._get_lexicon_option('identifier')
        rdtype = self._get_lexicon_option('type')
        name = (self._fqdn_name(self._get_lexicon_option('name'))
                if self._get_lexicon_option('name') else None)
        link = True if self._get_provider_option('linked') == 'yes' else False
        qname = name
        if identifier:
            rdtype, name, _ = self._parse_identifier(identifier)
        if action != 'list' and rdtype in ('A', 'AAAA', 'TXT') and name and link:
            if action != 'update' or name == qname or not qname:
                LOGGER.info('Hetzner => Enable CNAME lookup '
                            '(see --linked parameter)')
                return name, True
        LOGGER.info('Hetzner => Disable CNAME lookup '
                    '(see --linked parameter)')
        return name, False

    def _propagated_record(self, rdtype, name, content, nameservers=None):
        """
        If the publicly propagation check should be done, waits until the domain nameservers
        responses with the propagated record type, name & content and returns a boolean,
        if the publicly propagation was successful or not.
        """
        latency = self._get_provider_option('latency')
        propagated = True if self._get_provider_option('propagated') == 'yes' else False
        if propagated:
            retry, max_retry = 0, 20
            while retry < max_retry:
                for rdata in Provider._dns_lookup(name, rdtype, nameservers):
                    if content == rdata.to_text():
                        LOGGER.info('Hetzner => Record %s has %s %s', name, rdtype, content)
                        return True
                retry += 1
                retry_log = (', retry ({}/{}) in {}s...'.format((retry + 1), max_retry, latency)
                             if retry < max_retry else '')
                LOGGER.info('Hetzner => Record is not propagated%s', retry_log)
                time.sleep(latency)
        return False

    ###############################################################################
    # Hetzner API helpers
    ###############################################################################

    @staticmethod
    def _filter_dom(dom, filters, last_find_all=False):
        """
        If not exists, creates an DOM from a given session response, then filters the DOM
        via given API filters and returns the filtered DOM. The DOM is empty if the filters
        have no match.
        """
        if isinstance(dom, string_types):
            dom = BeautifulSoup(dom, 'html.parser')
        for idx, find in enumerate(filters, start=1):
            if not dom:
                break
            name, attrs = find.get('name'), find.get('attrs', {})
            if len(filters) == idx and last_find_all:
                dom = dom.find_all(name, attrs=attrs) if name else dom.find_all(attrs=attrs)
            else:
                dom = dom.find(name, attrs=attrs) if name else dom.find(attrs=attrs)
        return dom

    @staticmethod
    def _extract_hidden_data(dom):
        """
        Extracts hidden input data from DOM and returns the data as dictionary.
        """
        input_tags = dom.find_all('input', attrs={'type': 'hidden'})
        data = {}
        for input_tag in input_tags:
            data[input_tag['name']] = input_tag['value']
        return data

    @staticmethod
    def _extract_domain_id(string, regex):
        """
        Extracts domain ID from given string and returns the domain ID.
        """
        regex = re.compile(regex)
        match = regex.search(string)
        if not match:
            return False
        return str(match.group(1))

    @contextmanager
    def _session(self, domain, domain_id=None, get_zone=True):
        """
        Generates, authenticates and exits session to Hetzner account, and
        provides tuple of additional needed domain data (domain nameservers,
        zone and linked record name) to public methods. The tuple parameters
        are empty if not existent or specified. Exits session and raises error
        if provider fails during session.
        """
        name, link = self._link_record()
        qdomain, nameservers, cname = Provider._get_dns_cname(
            (name if name else domain + '.'), link)
        qdomain_id, zone = domain_id, None
        self.session = self._auth_session(self.username, self.password)
        try:
            if not domain_id or qdomain != domain:
                qdomain_id = self._get_domain_id(qdomain)
            if qdomain == domain:
                self.domain_id = qdomain_id
            if get_zone:
                zone = self._get_zone(qdomain, qdomain_id)
            yield {'nameservers': nameservers, 'zone': zone, 'cname': cname}
        except Exception as exc:
            raise exc
        finally:
            self._exit_session()

    def _auth_session(self, username, password):
        """
        Creates session to Hetzner account, authenticates with given credentials and
        returns the session, if authentication was successful. Otherwise raises error.
        """
        api = self.api[self.account]['auth']
        endpoint = api.get('endpoint', self.api[self.account]['endpoint'])
        session = requests.Session()
        session_retries = Retry(total=10, backoff_factor=0.5)
        session_adapter = requests.adapters.HTTPAdapter(max_retries=session_retries)
        session.mount('https://', session_adapter)
        response = session.request('GET', endpoint + api['GET'].get('url', '/'))
        dom = Provider._filter_dom(response.text, api['filter'])
        data = Provider._extract_hidden_data(dom)
        data[api['user']], data[api['pass']] = username, password
        response = session.request('POST', endpoint + api['POST']['url'], data=data)
        if Provider._filter_dom(response.text, api['filter']):
            LOGGER.error('Hetzner => Unable to authenticate session with %s account \'%s\': '
                         'Invalid credentials',
                         self.account, username)
            raise AssertionError
        LOGGER.info('Hetzner => Authenticate session with %s account \'%s\'',
                    self.account, username)
        return session

    def _exit_session(self):
        """
        Exits session to Hetzner account and returns.
        """
        api = self.api[self.account]
        response = self._get(api['exit']['GET']['url'])
        if not Provider._filter_dom(response.text, api['filter']):
            LOGGER.info('Hetzner => Exit session')
        else:
            LOGGER.warning('Hetzner => Unable to exit session')
        self.session = None
        return True

    def _get_domain_id(self, domain):
        """
        Pulls all domains managed by authenticated Hetzner account, extracts their IDs
        and returns the ID for the current domain, if exists. Otherwise raises error.
        """
        api = self.api[self.account]['domain_id']
        qdomain = dns.name.from_text(domain).to_unicode(True)
        domains, last_count, page = {}, -1, 0
        while last_count != len(domains):
            last_count = len(domains)
            page += 1
            url = (api['GET'].copy()).get('url', '/').replace('<index>', str(page))
            params = api['GET'].get('params', {}).copy()
            for param in params:
                params[param] = params[param].replace('<index>', str(page))
            response = self._get(url, query_params=params)
            domain_tags = Provider._filter_dom(response.text, api['filter'], True)
            for domain_tag in domain_tags:
                domain_id = Provider._extract_domain_id(dict(domain_tag.attrs)[api['id']['attr']],
                                                        api['id']['regex'])
                domain = (Provider._filter_dom(domain_tag, api['domain'])
                          .renderContents().decode('UTF-8'))
                domains[domain] = domain_id
                if domain == qdomain:
                    LOGGER.info('Hetzner => Get ID %s for domain %s', domain_id, qdomain)
                    return domain_id
        LOGGER.error('Hetzner => ID for domain %s does not exists', qdomain)
        raise AssertionError

    def _get_zone(self, domain, domain_id):
        """
        Pulls the zone for the current domain from authenticated Hetzner account and
        returns it as an zone object.
        """
        api = self.api[self.account]
        for request in api['zone']['GET']:
            url = (request.copy()).get('url', '/').replace('<id>', domain_id)
            params = request.get('params', {}).copy()
            for param in params:
                params[param] = params[param].replace('<id>', domain_id)
            response = self._get(url, query_params=params)
        dom = Provider._filter_dom(response.text, api['filter'])
        zone_file_filter = [{'name': 'textarea', 'attrs': {'name': api['zone']['file']}}]
        zone_file = Provider._filter_dom(dom, zone_file_filter).renderContents().decode('UTF-8')
        hidden = Provider._extract_hidden_data(dom)
        zone = {'data': dns.zone.from_text(zone_file, origin=domain, relativize=False),
                'hidden': hidden}
        LOGGER.info('Hetzner => Get zone for domain %s', domain)
        return zone

    def _post_zone(self, zone):
        """
        Pushes updated zone for current domain to authenticated Hetzner account and
        returns a boolean, if update was successful or not. Furthermore, waits until
        the zone has been taken over, if it is a Hetzner Robot account.
        """
        api = self.api[self.account]['zone']
        data = zone['hidden']
        data[api['file']] = zone['data'].to_text(relativize=True)
        response = self._post(api['POST']['url'], data=data)
        if Provider._filter_dom(response.text, api['filter']):
            LOGGER.error('Hetzner => Unable to update zone for domain %s: Syntax error\n\n%s',
                         zone['data'].origin.to_unicode(True),
                         zone['data'].to_text(relativize=True).decode('UTF-8'))
            return False

        LOGGER.info('Hetzner => Update zone for domain %s',
                    zone['data'].origin.to_unicode(True))
        if self.account == 'robot':
            latency = self._get_provider_option('latency')
            LOGGER.info('Hetzner => Wait %ds until Hetzner Robot has taken over zone...',
                        latency)
            time.sleep(latency)
        return True



"""Module provider for Linode"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['linode.com']


def provider_parser(subparser):
    """Module provider for Linode"""
    subparser.add_argument(
        "--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Linode"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.linode.com/api/'

    def _authenticate(self):
        self.domain_id = None
        payload = self._get('domain.list')
        for domain in payload['DATA']:
            if domain['DOMAIN'] == self.domain:
                self.domain_id = domain['DOMAINID']
        if self.domain_id is None:
            raise Exception('Domain not found')

    def _create_record(self, rtype, name, content):
        if not self._list_records(rtype, name, content):
            self._get('domain.resource.create', query_params={
                'DomainID': self.domain_id,
                'Name': self._relative_name(name),
                'Type': rtype,
                'Target': content,
                'TTL_sec': 0
            })

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get('domain.resource.list',
                            query_params={'DomainID': self.domain_id})
        resource_list = payload['DATA']
        if rtype:
            resource_list = [
                resource for resource in resource_list if resource['TYPE'] == rtype]
        if name:
            cmp_name = self._relative_name(name.lower())
            resource_list = [
                resource for resource in resource_list if resource['NAME'] == cmp_name]
        if content:
            resource_list = [
                resource for resource in resource_list if resource['TARGET'] == content]

        processed_records = []
        for resource in resource_list:
            processed_records.append({
                'id': resource['RESOURCEID'],
                'type': resource['TYPE'],
                'name': self._full_name(resource['NAME']),
                'ttl': resource['TTL_SEC'],
                'content': resource['TARGET']
            })
        LOGGER.debug('list_records: %s', processed_records)
        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            resources = self._list_records(rtype, name, None)
            identifier = resources[0]['id'] if resources else None

        LOGGER.debug('update_record: %s', identifier)

        self._get('domain.resource.update', query_params={
            'DomainID': self.domain_id,
            'ResourceID': identifier,
            'Name': self._relative_name(name).lower() if name else None,
            'Type': rtype if rtype else None,
            'Target': content if content else None
        })

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self._list_records(rtype, name, content)
            delete_resource_id = [resource['id'] for resource in resources]
        else:
            delete_resource_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_resource_id)

        for resource_id in delete_resource_id:
            self._get('domain.resource.delete', query_params={
                'DomainID': self.domain_id,
                'ResourceID': resource_id
            })

        return True

    # Helpers
    def _request(self, action='GET', url='', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

        query_params['api_key'] = self._get_provider_option('auth_token')
        query_params['resultFormat'] = 'JSON'
        query_params['api_action'] = url

        response = requests.request(action, self.api_endpoint, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == 'DELETE':
            return ''
        result = response.json()
        if result['ERRORARRAY']:
            raise Exception('Linode api error: {0}'.format(result['ERRORARRAY']))
        return result



"""Module provider for Namesilo"""
from __future__ import absolute_import
import logging
from xml.etree import ElementTree

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['namesilo.com']


def provider_parser(subparser):
    """Configure provider parser for Namesilo"""
    subparser.add_argument(
        "--auth-token", help="specify key for authentication")


class Provider(BaseProvider):
    """Provider class for Namesilo"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = self._get_provider_option(
            'api_endpoint') or 'https://www.namesilo.com/api'

    def _authenticate(self):
        self._get('/getDomainInfo', {'domain': self.domain})
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            'domain': self.domain_id,
            'rrhost': self._relative_name(name),
            'rrtype': rtype,
            'rrvalue': content
        }
        if self._get_lexicon_option('ttl'):
            record['rrttl'] = self._get_lexicon_option('ttl')
        try:
            self._get('/dnsAddRecord', record)
        except ValueError as err:
            # noop if attempting to create record that already exists.
            LOGGER.debug('Ignoring error: %s', err)

        LOGGER.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        query = {'domain': self.domain_id}

        payload = self._get('/dnsListRecords', query)
        records = []
        for record in payload.find('reply').findall('resource_record'):
            processed_record = {
                'type': record.find('type').text,
                'name': record.find('host').text,
                'ttl': record.find('ttl').text,
                'content': record.find('value').text,
                'id': record.find('record_id').text
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record['type'] == rtype]
        if name:
            records = [record for record in records if record['name']
                       == self._full_name(name)]
        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {
            'domain': self.domain_id,
            'rrid': identifier
        }
        # if rtype:
        #     data['type'] = rtype
        if name:
            data['rrhost'] = self._relative_name(name)
        if content:
            data['rrvalue'] = content
        if self._get_lexicon_option('ttl'):
            data['rrttl'] = self._get_lexicon_option('ttl')

        self._get('/dnsUpdateRecord', data)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        data = {
            'domain': self.domain_id
        }

        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            data['rrid'] = record_id
            self._get('/dnsDeleteRecord', data)

        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        query_params['version'] = 1
        query_params['type'] = 'xml'
        query_params['key'] = self._get_provider_option('auth_token')
        response = requests.request(action, self.api_endpoint +
                                    url, params=query_params)
        # data=json.dumps(data))
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        tree = ElementTree.ElementTree(ElementTree.fromstring(response.content))
        root = tree.getroot()
        if root.find('reply').find('code').text == '280':
            raise ValueError('An error occurred: {0}, {1}'.format(
                root.find('reply').find('detail').text, root.find('reply').find('code').text))
        elif root.find('reply').find('code').text != '300':
            raise Exception('An error occurred: {0}, {1}'.format(
                root.find('reply').find('detail').text, root.find('reply').find('code').text))

        return root



"""Module provider for exoscale"""
from __future__ import absolute_import
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

HOUR = 3600

NAMESERVER_DOMAINS = ['exoscale.ch']


def provider_parser(subparser):
    """Generate subparser for exoscale"""
    subparser.add_argument(
        "--auth-key", help="specify API key for authentication"
    )
    subparser.add_argument(
        "--auth-secret", help="specify API secret for authentication"
    )


class Provider(BaseProvider):
    """Provider class for exoscale"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = 'https://api.exoscale.ch/dns'

    def _authenticate(self):
        """An innocent call to check that the credentials are okay."""
        response = self._get("/v1/domains/{0}".format(self.domain))

        self.domain_id = response["domain"]["id"]

    def _create_record(self, rtype, name, content):
        """Create record if doesnt already exist with same content"""
        # check if record already exists
        existing_records = self._list_records(rtype, name, content)
        if len(existing_records) >= 1:
            return True

        record = {
            "record_type": rtype,
            "name": self._relative_name(name),
            "content": content,
        }
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            record["prio"] = self._get_lexicon_option("priority")

        payload = self._post(
            "/v1/domains/{0}/records".format(self.domain),
            {"record": record},
        )

        status = "id" in payload.get("record", {})
        LOGGER.debug("create_record: %s", status)
        return status

    def _list_records(self, rtype=None, name=None, content=None):
        """List all records.

        record_type, name and content are used to filter the records.
        If possible it filters during the query, otherwise afterwards.
        An empty list is returned if no records are found.
        """

        filter_query = {}
        if rtype:
            filter_query["record_type"] = rtype
        if name:
            name = self._relative_name(name)
            filter_query["name"] = name
        payload = self._get(
            "/v1/domains/{0}/records".format(self.domain),
            query_params=filter_query,
        )

        records = []
        for data in payload:
            record = data["record"]

            if content and record["content"] != content:
                continue

            if record["name"] == "":
                rname = self.domain
            else:
                rname = ".".join((record["name"], self.domain))

            processed_record = {
                "type": record["record_type"],
                "name": rname,
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            if record["prio"]:
                processed_record["options"] = {
                    "mx": {"priority": record["prio"]}
                }
            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """Create or update a record."""
        record = {}

        if not identifier:
            records = self._list_records(rtype, name, content)
            identifiers = [r["id"] for r in records]
        else:
            identifiers = [identifier]

        if name:
            record["name"] = self._relative_name(name)
        if content:
            record["content"] = content
        if self._get_lexicon_option('ttl'):
            record["ttl"] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            record["prio"] = self._get_lexicon_option('priority')

        LOGGER.debug("update_records: %s", identifiers)

        for record_id in identifiers:
            self._put(
                "/v1/domains/{0}/records/{1}".format(
                    self.domain, identifier
                ),
                record,
            )
            LOGGER.debug("update_record: %s", record_id)

        LOGGER.debug("update_record: %s", True)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Delete an existing record.

        If the record doesn't exist, does nothing.
        """
        if not identifier:
            records = self._list_records(rtype, name, content)
            identifiers = [record["id"] for record in records]
        else:
            identifiers = [identifier]

        LOGGER.debug("delete_records: %s", identifiers)

        for record_id in identifiers:
            self._delete(
                "/v1/domains/{0}/records/{1}".format(
                    self.domain, record_id
                )
            )
            LOGGER.debug("delete_record: %s", record_id)

        LOGGER.debug("delete_record: %s", True)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        """Performs the request to the API"""
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {"Accept": "application/json"}

        default_headers["X-DNS-Token"] = ":".join(
            (self._get_provider_option("auth_key"),
             self._get_provider_option("auth_secret"))
        )

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            json=data,
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if response.text and response.json() is None:
            raise Exception("No data returned")

        return response.json() if response.text else None



"""Integration tests for Subreg"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.subreg import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class SubregProviderTests(TestCase, IntegrationTests):
    """TestCase for Subreg"""
    Provider = Provider
    provider_name = 'subreg'
    domain = 'oldium.xyz'



"""
The internetbs api has an undocumented rate-limit ; here is what the support could tell me about
it: "Please note that our API is intended mainly for domain name registrations and that's why we
offer it for free and your registration volume has to stay in phase with the requests you perform.
If you keep a reasonable ratio between check and successful registrations, you should never hit
the limits. Here are the limits:
- per minute = 60+N
- per hour = 500+N
- per day = 1000+2*N
- per week = 2000+2.5*N
- per month = 3000+3*N
Where N is the number of registered domains in your account (so excluding pending
transfers/trades)."
"""
from __future__ import absolute_import
import hashlib
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['topdns.com']


def provider_parser(subparser):
    """Configure provider parser for internetbs"""
    subparser.add_argument(
        "--auth-key", help="specify API key for authentication")
    subparser.add_argument(
        "--auth-password", help="specify password for authentication")


class Provider(BaseProvider):
    """Provider class for internetbs"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.internet.bs'

    def _authenticate(self):
        payload = self._get('/Domain/List', {
            'searchTermFilter': self.domain
        })

        LOGGER.debug('authenticate debug: %s', payload)
        if not payload['status']:
            raise Exception('Internal error. This should not happen')
        if payload['status'] != 'SUCCESS':
            raise Exception('Api error: {0}'.format(payload['message']))
        if self.domain not in payload['domain']:
            raise Exception('Domain not found')
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        # Skip execution if such a record already exists
        existing_records = self._list_records(rtype, name, content)
        if existing_records:
            return True

        query = {'Type': rtype, 'FullRecordName': self._full_name(
            name), 'Value': content}
        ttl = self._get_lexicon_option('ttl')
        if ttl:
            query['Ttl'] = ttl

        # for MX records, query['Priority'] could be set (default is 10)

        payload = self._post('/Domain/DnsRecord/Add', None, query)
        LOGGER.debug('authenticate debug: %s', payload)
        if not payload['status']:
            raise Exception('Internal error. This should not happen')
        if payload['status'] != 'SUCCESS':
            raise Exception('Api error: {0}'.format(payload['message']))
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        query = {'Domain': self.domain}
        if rtype:
            query['FilterType'] = rtype

        payload = self._get('/Domain/DnsRecord/List', query)

        record_list = payload['records']
        if name:
            cmp_name = self._relative_name(name.lower())
            record_list = [record for record in record_list
                           if self._relative_name(record['name']) == cmp_name]
        if content:
            record_list = [
                record for record in record_list if record['value'] == content]

        records = []
        for record in record_list:
            processed_record = {
                'type': record['type'],
                'name': record['name'],
                'ttl': record['ttl'],
                'content': record['value'],
                # internetbs api does not provide id
                'id': hashlib.sha1('/'.join([record['type'],
                                             record['name'],
                                             record['value']]).encode('utf-8')).hexdigest()
            }
            records.append(processed_record)

        LOGGER.debug('list_records: %s', records)
        return records

    # Update a record.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            to_update = next(
                (r for r in records if r['id'] == identifier), None)
            query = {'Type': to_update['type'],
                     'FullRecordName': to_update['name'],
                     'NewValue': content}
        else:
            query = {'Type': rtype, 'FullRecordName': self._full_name(
                name), 'NewValue': content}
        LOGGER.debug('update_record query: %s', query)
        payload = self._post('/Domain/DnsRecord/Update', None, query)
        LOGGER.debug('update_record payload: %s', payload)

        if not payload['status']:
            raise Exception('Internal error. This should not happen')
        if payload['status'] != 'SUCCESS':
            raise Exception('Api error: {0}'.format(payload['message']))
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            to_update = next(
                (r for r in records if r['id'] == identifier), None)
            if not to_update:
                return True
            query = {'Type': to_update['type'],
                     'FullRecordName': to_update['name'],
                     'NewValue': content}
        else:
            query = {'Type': rtype,
                     'FullRecordName': self._full_name(name)}
            if content:
                query['Value'] = content
                if not self._list_records(rtype, name, content):
                    return True
            else:
                if not self._list_records(rtype, name):
                    return True

        LOGGER.debug('delete_record query: %s', query)
        payload = self._post('/Domain/DnsRecord/Remove', None, query)
        LOGGER.debug('delete_record payload: %s', payload)

        if not payload['status']:
            raise Exception('Internal error. This should not happen')
        if payload['status'] != 'SUCCESS':
            raise Exception('Api error: {0}'.format(payload['message']))
        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        query_params['ResponseFormat'] = 'json'
        query_params['ApiKey'] = self._get_provider_option('auth_key')
        query_params['Password'] = self._get_provider_option('auth_password')
        request = requests.request(action, self.api_endpoint + url, params=query_params,
                                   data=json.dumps(data),
                                   headers={'Content-Type': 'application/json'})
        # if the request fails for any reason, throw an error.
        request.raise_for_status()
        return request.json()



"""Module provider for Easyname DNS"""
from __future__ import absolute_import, print_function
import logging

from requests import Response, Session

from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['easyname.eu']


def provider_parser(subparser):
    """Configure provider parser for Easyname DNS"""
    subparser.description = """A provider for Easyname DNS."""
    subparser.add_argument(
        '--auth-username',
        help='Specify username used to authenticate'
    )
    subparser.add_argument(
        '--auth-password',
        help='Specify password used to authenticate',
    )


class Provider(BaseProvider):
    """
        easyname provider
    """

    URLS = {
        'login': 'https://my.easyname.com/en/login',
        'domain_list': 'https://my.easyname.com/domains',
        'overview': 'https://my.easyname.com/hosting/view-user.php',
        'dns': 'https://my.easyname.com/domains/settings/dns.php?domain={}',
        'dns_create_entry': 'https://my.easyname.com/domains/settings/form.php?domain={}',
        'dns_delete_entry':
            'https://my.easyname.com/domains/settings/delete_record.php?domain={}&confirm=1&id={}'
    }

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.session = Session()
        self.domain_id = None
        self._records = None

    def _authenticate(self):
        """
        Authenticates against Easyname website and try to find out the domain
        id.
        Easyname uses a CSRF token in its login form, so two requests are
        neccessary to actually login.

        Returns:
          bool: True if domain id was found.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
          ValueError: When login data is wrong or the domain does not exist.
        """
        csrf_token = self._get_csrf_token()
        self._login(csrf_token)

        domain_text_element = self._get_domain_text_of_authoritative_zone()
        self.domain_id = self._get_domain_id(domain_text_element)
        LOGGER.debug('Easyname domain ID: %s', self.domain_id)

        return True

    def _create_record(self, rtype, name, content):
        return self._create_record_internal(rtype=rtype, name=name, content=content)

    def _create_record_internal(self, rtype, name, content, identifier=None):
        """
        Create a new DNS entry in the domain zone if it does not already exist.

        Args:
          rtype (str): The DNS type (e.g. A, TXT, MX, etc) of the new entry.
          name (str): The name of the new DNS entry, e.g the domain for which a
                      MX entry shall be valid.
          content (str): The content of the new DNS entry, e.g. the mail server
                         hostname for a MX entry.
          [identifier] (str): The easyname id of a DNS entry. Use to overwrite an
                    existing entry.

        Returns:
          bool: True if the record was created successfully, False otherwise.
        """
        name = self._relative_name(name) if name is not None else name
        LOGGER.debug('Creating record with name %s', name)
        if self._is_duplicate_record(rtype, name, content):
            return True

        data = self._get_post_data_to_create_dns_entry(rtype, name, content, identifier)
        LOGGER.debug('Create DNS data: %s', data)
        create_response = self.session.post(
            self.URLS['dns_create_entry'].format(self.domain_id),
            data=data
        )
        self._invalidate_records_cache()
        self._log('Create DNS entry', create_response)

        # Pull a list of records and check for ours
        was_success = len(self._list_records(rtype, name, content)) > 0
        if was_success:
            msg = 'Successfully added record %s'
        else:
            msg = 'Failed to add record %s'

        LOGGER.info(msg, name)
        return was_success

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete one or more DNS entries in the domain zone that match the given
        criteria.

        Args:
          [identifier] (str): An ID to match against DNS entry easyname IDs.
          [rtype] (str): A DNS rtype (e.g. A, TXT, MX, etc) to match against DNS
                      entry types.
          [name] (str): A name to match against DNS entry names.
          [content] (str): A content to match against a DNS entry contents.

        Returns:
          bool: True if the record(s) were deleted successfully, False
                otherwise.
        """
        success_url = self.URLS['dns'].format(self.domain_id)
        record_ids = self._get_matching_dns_entry_ids(identifier, rtype,
                                                      name, content)
        LOGGER.debug('Record IDs to delete: %s', record_ids)

        success = True
        for rec_id in record_ids:
            delete_response = self.session.get(
                self.URLS['dns_delete_entry'].format(self.domain_id, rec_id))
            self._invalidate_records_cache()
            self._log('Delete DNS entry {}'.format(rec_id), delete_response)
            success = success and delete_response.url == success_url

        return success

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Update a DNS entry identified by identifier or name in the domain zone.
        Any non given argument will leave the current value of the DNS entry.

        Args:
          identifier (str): The easyname id of the DNS entry to update.
          [rtype] (str): The DNS rtype (e.g. A, TXT, MX, etc) of the new entry.
          [name] (str): The name of the new DNS entry, e.g the domain for which
                        a MX entry shall be valid.
          [content] (str): The content of the new DNS entry, e.g. the mail
                           server hostname for a MX entry.

        Returns:
          bool: True if the record was updated successfully, False otherwise.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        if identifier is not None:
            identifier = int(identifier)
            records = self._list_records_internal(identifier=identifier)
        else:
            records = self._list_records_internal(name=name, rtype=rtype)
        LOGGER.debug('Records to update (%d): %s', len(records), records)
        assert records, 'No record found to update'
        success = True

        for record in records:
            name = name if name is not None else record['name']
            rtype = rtype if rtype is not None else record['type']
            content = content if content is not None \
                else record['content']
            success = success and self._create_record_internal(
                rtype, name, content, record['id'])
        return success

    def _list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(self, rtype=None, name=None, content=None, identifier=None):
        """
        Filter and list DNS entries of domain zone on Easyname.
        Easyname shows each entry in a HTML table row and each attribute on a
        table column.

        Args:
          [rtype] (str): Filter by DNS rtype (e.g. A, TXT, MX, etc)
          [name] (str): Filter by the name of the DNS entry, e.g the domain for
                      which a MX entry shall be valid.
          [content] (str): Filter by the content of the DNS entry, e.g. the
                           mail server hostname for a MX entry.
          [identifier] (str): Filter by the easyname id of the DNS entry.

        Returns:
          list: A list of DNS entries. A DNS entry is an object with DNS
                attribute names as keys (e.g. name, content, priority, etc)
                and additionally an id.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        name = self._full_name(name) if name is not None else name
        if self._records is None:
            records = []
            rows = self._get_dns_entry_trs()

            for index, row in enumerate(rows):
                self._log('DNS list entry', row)
                try:
                    rec = {}
                    if row.has_attr('ondblclick'):
                        rec['id'] = int(row['ondblclick'].split(
                            'id=')[1].split("'")[0])
                    else:
                        rec['id'] = -index

                    columns = row.find_all('td')
                    rec['name'] = (columns[0].string or '').strip()
                    rec['type'] = (columns[1].contents[1] or '').strip()
                    rec['content'] = (columns[2].string or '').strip()
                    rec['priority'] = (columns[3].string or '').strip()
                    rec['ttl'] = (columns[4].string or '').strip()

                    if rec['priority']:
                        rec['priority'] = int(rec['priority'])

                    if rec['ttl']:
                        rec['ttl'] = int(rec['ttl'])
                except Exception as error:
                    errmsg = 'Cannot parse DNS entry ({}).'.format(error)
                    LOGGER.warning(errmsg)
                    raise AssertionError(errmsg)
                records.append(rec)
            self._records = records

        records = self._filter_records(self._records, rtype, name, content, identifier)
        LOGGER.debug('Final records (%d): %s', len(records), records)
        return records

    def _request(self, action='GET', url='/', data=None, query_params=None):
        pass

    def _invalidate_records_cache(self):
        """
        Invalidate DNS entries cache such that list_records will do a new
        request to retrieve DNS entries.
        """
        self._records = None

    def _get_post_data_to_create_dns_entry(self, rtype, name, content, identifier=None):
        """
        Build and return the post date that is needed to create a DNS entry.
        """
        is_update = identifier is not None
        if is_update:
            records = self._list_records_internal(identifier=identifier)
            assert len(records) == 1, 'ID is not unique or does not exist'
            record = records[0]
            LOGGER.debug('Create post data to update record: %s', record)

        data = {
            'id': str(identifier) if is_update else '',
            'action': 'save',
            'name': name,
            'type': rtype,
            'content': content,
            'prio': str(record['priority']) if is_update else '10',
            'ttl': str(record['ttl']) if is_update else '360',
            'commit': ''
        }
        ttl = self._get_lexicon_option('ttl')
        if ttl and ttl > 360:
            data['ttl'] = str(ttl)

        prio = self._get_lexicon_option('priority')
        if prio and prio > 0:
            data['prio'] = str(prio)

        return data

    def _is_duplicate_record(self, rtype, name, content):
        """Check if DNS entry already exists."""
        records = self._list_records(rtype, name, content)
        is_duplicate = len(records) >= 1
        if is_duplicate:
            LOGGER.info('Duplicate record %s %s %s, NOOP', rtype, name, content)
        return is_duplicate

    def _get_matching_dns_entry_ids(self, identifier=None, rtype=None,
                                    name=None, content=None):
        """Return a list of DNS entries that match the given criteria."""
        record_ids = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            record_ids = [record['id'] for record in records]
        else:
            record_ids.append(identifier)
        return record_ids

    def _get_dns_entry_trs(self):
        """
        Return the TR elements holding the DNS entries.
        """
        from bs4 import BeautifulSoup
        dns_list_response = self.session.get(
            self.URLS['dns'].format(self.domain_id))
        self._log('DNS list', dns_list_response)
        assert dns_list_response.status_code == 200, \
            'Could not load DNS entries.'

        html = BeautifulSoup(dns_list_response.content, 'html.parser')
        self._log('DNS list', html)
        dns_table = html.find('table', {'id': 'cp_domains_dnseintraege'})
        assert dns_table is not None, 'Could not find DNS entry table'

        def _is_zone_tr(elm):
            has_ondblclick = elm.has_attr('ondblclick')
            has_class = elm.has_attr('class')
            return elm.name.lower() == 'tr' and (has_class or has_ondblclick)

        rows = dns_table.findAll(_is_zone_tr)
        assert rows is not None and rows, 'Could not find any DNS entries'
        return rows

    def _filter_records(self, records, rtype=None, name=None, content=None, identifier=None):  # pylint: disable=too-many-arguments,no-self-use
        """
        Filter dns entries based on type, name or content.
        """
        if not records:
            return []
        if identifier is not None:
            LOGGER.debug('Filtering %d records by id: %s', len(records), identifier)
            records = [record for record in records if record['id'] == identifier]
        if rtype is not None:
            LOGGER.debug('Filtering %d records by type: %s', len(records), rtype)
            records = [record for record in records if record['type'] == rtype]
        if name is not None:
            LOGGER.debug('Filtering %d records by name: %s', len(records), name)
            if name.endswith('.'):
                name = name[:-1]
            records = [record for record in records if name == record['name']]
        if content is not None:
            LOGGER.debug('Filtering %d records by content: %s', len(records), content.lower())
            records = [record for record in records if
                       record['content'].lower() == content.lower()]
        return records

    def _get_csrf_token(self):
        """Return the CSRF Token of easyname login form."""
        from bs4 import BeautifulSoup
        home_response = self.session.get(self.URLS['login'])
        self._log('Home', home_response)
        assert home_response.status_code == 200, \
            'Could not load Easyname login page.'

        html = BeautifulSoup(home_response.content, 'html.parser')
        self._log('Home', html)
        csrf_token_field = html.find('input', {'id': 'loginxtoken'})
        assert csrf_token_field is not None, 'Could not find login token.'
        return csrf_token_field['value']

    def _login(self, csrf_token):
        """Attempt to login session on easyname."""
        login_response = self.session.post(
            self.URLS['login'],
            data={
                'username': self._get_provider_option('auth_username') or '',
                'password': self._get_provider_option('auth_password') or '',
                'submit': '',
                'loginxtoken': csrf_token,
            }
        )
        self._log('Login', login_response)
        assert login_response.status_code == 200, \
            'Could not login due to a network error.'
        assert login_response.url == self.URLS['overview'], \
            'Easyname login failed, bad EASYNAME_USER or EASYNAME_PASS.'

    def _get_domain_text_of_authoritative_zone(self):
        """Get the authoritative name zone."""
        # We are logged in, so get the domain list
        from bs4 import BeautifulSoup
        zones_response = self.session.get(self.URLS['domain_list'])
        self._log('Zone', zones_response)
        assert zones_response.status_code == 200, \
            'Could not retrieve domain list due to a network error.'

        html = BeautifulSoup(zones_response.content, 'html.parser')
        self._log('Zone', html)
        domain_table = html.find('table', {'id': 'cp_domain_table'})
        assert domain_table is not None, 'Could not find domain table'

        # (Sub)domains can either be managed in their own zones or by the
        # zones of their parent (sub)domains. Iterate over all subdomains
        # (starting with the deepest one) and see if there is an own zone
        # for it.
        domain = self.domain or ''
        domain_text = None
        subdomains = domain.split('.')
        while True:
            domain = '.'.join(subdomains)
            LOGGER.debug('Check if %s has own zone', domain)
            domain_text = domain_table.find(string=domain)
            if domain_text is not None or len(subdomains) < 3:
                break
            subdomains.pop(0)

        # Update domain to equal the zone's domain. This is important if we are
        # handling a subdomain that has no zone of itself. If we do not do
        # this, self._relative_name will strip also a part of the subdomain
        # away.
        self.domain = domain
        assert domain_text is not None, \
            'The domain does not exist on Easyname.'
        return domain_text

    def _get_domain_id(self, domain_text_element):  # pylint: disable=no-self-use
        """Return the easyname id of the domain."""
        try:
            # Hierarchy: TR > TD > SPAN > Domain Text
            tr_anchor = domain_text_element.parent.parent.parent
            td_anchor = tr_anchor.find('td', {'class': 'td_2'})
            link = td_anchor.find('a')['href']
            domain_id = link.rsplit('/', 1)[-1]
            return domain_id
        except Exception as error:
            errmsg = ('Cannot get the domain id even though the domain seems '
                      'to exist (%s).', error)
            LOGGER.warning(errmsg)
            raise AssertionError(errmsg)

    def _log(self, name, element):  # pylint: disable=no-self-use
        """
        Log Response and Tag elements. Do nothing if elements is none of them.
        """
        from bs4 import BeautifulSoup, Tag
        if isinstance(element, Response):
            LOGGER.debug('%s response: URL=%s Code=%s', name, element.url, element.status_code)
        elif isinstance(element, (BeautifulSoup, Tag)):
            LOGGER.debug('%s HTML:\n%s', name, element)



"""Integration tests for DNSMadeEasy"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.dnsmadeeasy import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class DnsmadeeasyProviderTests(TestCase, IntegrationTests):
    """TestCase for DNSMadeEasy"""
    Provider = Provider
    provider_name = 'dnsmadeeasy'
    domain = 'capsulecd.com'

    def _test_parameters_overrides(self):
        return {'api_endpoint': 'http://api.sandbox.dnsmadeeasy.com/V2.0'}

    def _filter_headers(self):
        return ['x-dnsme-apiKey', 'x-dnsme-hmac', 'Authorization']

    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return



"""Integration tests for Softlayer"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.softlayer import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class SoftLayerProviderTests(TestCase, IntegrationTests):
    """TestCase for SoftLayer"""
    Provider = Provider
    provider_name = 'softlayer'
    domain = 'example.com'

    # SoftLayer does not provide a sandbox API; actual credentials are required
    # Keeping this here for when fixtures need to be regenerated
    # def _test_options(self):
    #    options = super(SoftLayerProviderTests, self)._test_options()
    #    options.update({
    #        'auth_username': 'foo',
    #        'auth_api_key': 'bar'
    #        })
    #    return options

    # TODO: the following skipped suite and fixtures should be enabled
    @pytest.fixture(autouse=True)
    def _skip_suite(self, request):  # pylint: disable=no-self-use
        if request.node.get_closest_marker('ext_suite_1'):
            pytest.skip('Skipping extended suite')



"""Integration tests for Localzone"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.localzone import Provider

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class LocalzoneProviderTests(TestCase, IntegrationTests):
    """Integration tests for Localzone"""
    Provider = Provider
    provider_name = "localzone"
    domain = "example.com"
    file_uri = "https://raw.githubusercontent.com/ags-slc/localzone/master/tests/zonefiles/db.example.com"  # pylint: disable=line-too-long
    filename, headers = urlretrieve(file_uri)

    def _test_parameters_overrides(self):
        options = {
            "filename": self.filename
        }

        return options

    def _test_fallback_fn(self):
        return lambda _: None

    @pytest.mark.skip(reason="localzone does not require authentication")
    def test_provider_authenticate(self):
        return

    @pytest.mark.skip(reason="localzone does not require authentication")
    def test_provider_authenticate_with_unmanaged_domain_should_fail(self):
        return
